import logging

import kerberos

from django import http
from django.template.response import TemplateResponse
from django.conf import settings
from django.views.generic.base import View
from django.contrib import messages

from django.contrib.auth import authenticate, login as auth_login

from . import app_settings


class NegotiateView(View):
    NEXT_URL_FIELD = 'next'
    unauthorized_template_name = 'django_kerberos/unauthorized.html'
    error_template_name = 'django_kerberos/error.html'

    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger(__name__)
        super(NegotiateView, self).__init__(*args, **kwargs)

    def challenge(self, request, *args, **kwargs):
        '''Send negotiate challenge'''
        response = TemplateResponse(request, self.unauthorized_template_name, status=401)
        response['WWW-Authenticate'] = 'Negotiate'
        return response

    def host(self, request):
        return app_settings.HOSTNAME or request.get_host().split(':')[0]

    def principal_valid(self, request, *args, **kwargs):
        '''Do something with the principal we received'''
        self.logger.info(u'got ticket for principal %s', self.principal)
        user = authenticate(principal=self.principal)
        next_url = request.REQUEST.get(self.NEXT_URL_FIELD) or settings.LOGIN_REDIRECT_URL
        if user:
            self.login_user(request, user)
        if request.is_ajax():
            return http.HttpResponse('true' if user else 'false', content_type='application/json')
        else:
            if not user:
                self.logger.warning(u'principal %s has no local user', self.principal)
                messages.warning(request, _('Principal %s could not be authenticated') %
                                 self.principal)
            return http.HttpResponseRedirect(next_url)

    def login_user(self, request, user):
        auth_login(request, user)

    def negotiate(self, request, *args, **kwargs):
        '''Try to authenticate the user using SPNEGO and Kerberos'''

        if 'HTTP_AUTHORIZATION' in request.META:
            kind, authstr = request.META['HTTP_AUTHORIZATION'].split(' ', 1)
            if kind == 'Negotiate':
                service = 'HTTP@%s' % self.host(request)
                self.logger.debug(u'using service name %s', service)
                self.logger.debug(u'Negotiate authstr %r', authstr)
                try:
                    result, context = kerberos.authGSSServerInit(service)
                except kerberos.KrbError, e:
                    self.logger.warning(u'exception during authGSSServerInit: %s, certainly a '
                                        u'keytab problem', e)
                    details = (u'exception during authGSSServerInit: %s, certainly a '
                               u'keytab problem' % e)
                    return TemplateResponse(request, self.error_template_name,
                                            context={'details': details}, status=500)
                # ensure context is finalized
                try:
                    if result != 1:
                        self.logger.warning(u'authGSSServerInit result is non-zero: %s', result)
                        details = u'authGSSServerInit result is non-zero: %s' % result
                        return TemplateResponse(request, self.error_template_name,
                                                context={'details': details}, status=500)
                    try:
                        r = kerberos.authGSSServerStep(context, authstr)
                    except kerberos.KrbError, e:
                        self.logger.warning(u'exception during authGSSServerStep: %s', e)
                        details = u'exception during authGSSServerStep: %s' % e
                        return TemplateResponse(request, self.error_template_name,
                                                context={'details': details}, status=500)
                    if r == 1:
                        gssstring = kerberos.authGSSServerResponse(context)
                    else:
                        return self.challenge(request, *args, **kwargs)
                    try:
                        self.principal = kerberos.authGSSServerUserName(context)
                    except kerberos.KrbError, e:
                        self.logger.warning(u'exception during authGSSServerUserName: %s', e)
                        details = u'exception during authGSSServerUserName: %s' % e
                        return TemplateResponse(request, self.error_template_name,
                                                context={'details': details}, status=500)
                finally:
                    kerberos.authGSSServerClean(context)
                response = self.principal_valid(request, *args, **kwargs)
                if response:
                    response['WWW-Authenticate'] = 'Negotiate %s' % gssstring
                    return response
        return self.challenge(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        return self.negotiate(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.negotiate(request, *args, **kwargs)

login = NegotiateView.as_view()



import logging

from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import force_bytes

from django.contrib.auth.hashers import BasePasswordHasher

import kerberos

from . import app_settings

class KerberosHasher(BasePasswordHasher):
    '''A pseudo hasher which just validate that the given password
       match a given Kerberos identity'''
    algorithm = 'kerberos'

    def default_realm(self):
        '''Default realm for usernames without a realm'''
        return app_settings.DEFAULT_REALM

    def service_principal(self):
        if not app_settings.SERVICE_PRINCIPAL:
            raise ImproperlyConfigured('Kerberos pseudo password hasher needs '
                    'the setting KERBEROS_SERVICE_PRINCIPAL to be '
                    'set')
        return app_settings.SERVICE_PRINCIPAL

    def verify(self, password, encoded):
        algorithm, principal = encoded.split('$', 2)
        assert algorithm == self.algorithm
        principal = force_bytes(principal)
        password = force_bytes(password)
        try:
            return kerberos.checkPassword(principal, password,
                    self.service_principal(), self.default_realm())
        except kerberos.KrbError, e:
            logging.getLogger(__name__).error('password validation'
                    'for principal %r failed %s', principal, e)
            return False





import pytest

import kerberos


@pytest.fixture
def kerberos_mock(request, mocker):
    a = {}
    d = (
        'kerberos.authGSSServerInit',
        'kerberos.authGSSServerStep',
        'kerberos.authGSSServerResponse',
        'kerberos.authGSSServerUserName',
        'kerberos.authGSSServerClean'
    )
    for name in d:
        if hasattr(request, 'param') and name in request.param:
            continue
        a[name] = mocker.patch(name)
    return a


def test_login_no_header(client, settings, kerberos_mock):
    client.get('/login/')
    for mock in kerberos_mock.values():
        assert mock.call_count == 0


@pytest.mark.parametrize('kerberos_mock', ['kerberos.authGSSServerInit'], indirect=True)
def test_login_missing_keytab(client, settings, kerberos_mock, caplog):
    resp = client.get('/login/', HTTP_AUTHORIZATION='Negotiate coin')
    for key, mock in kerberos_mock.iteritems():
        assert mock.call_count == 0
    assert 'keytab problem' in resp.content
    assert 'keytab problem' in caplog.text()





import re
import logging

from . import app_settings

from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import force_bytes

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

import kerberos

class KerberosBackend(ModelBackend):
    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def username_from_principal(self, principal):
        '''Make a username from a principal name'''
        username, domain = principal.rsplit('@', 1)
        return u'{0}@{1}'.format(username, domain.lower())

    def authorize_principal(self, principal):
        '''Is this principal authorized to login ?'''
        return True

    def provision_user(self, principal, user):
        '''Modify user based on information we can retrieve on this principal'''
        if app_settings.BACKEND_ADMIN_REGEXP:
            if re.match(app_settings.BACKEND_ADMIN_REGEXP, principal):
                if not user.is_staff or not user.is_superuser:
                    self.logger.info('giving superuser power to principal %r', principal)
                    user.is_staff = True
                    user.is_superuser = True
                    user.save()

    def should_create_user(self):
        '''Should we create users for new principals ?'''
        return app_settings.BACKEND_CREATE

    def lookup_user(self, principal):
        '''Find the user model linked to this principal'''
        User = get_user_model()
        username_field = getattr(User, 'USERNAME_FIELD', 'username')
        username = self.username_from_principal(principal)
        kwargs = {username_field: username}
        if self.should_create_user():
            user, created = User.objects.get_or_create(**kwargs)
            if created:
                user.set_unusable_password()
                user.save()
        else:
            try:
                user = User.objects.get(**kwargs)
            except User.DoesNotExist:
                return
        self.provision_user(principal, user)
        return user


    def authenticate(self, principal=None, **kwargs):
        if principal and self.authorize_principal(principal):
            return self.lookup_user(principal)


class KerberosPasswordBackend(KerberosBackend):
    def default_realm(self):
        '''Default realm for usernames without a realm'''
        return app_settings.DEFAULT_REALM

    def principal_from_username(self, username):
        realm = self.default_realm()
        if '@' not in username and realm:
            username = u'%s@%s' % (username, realm)
        return username

    def keep_password(self):
        '''Do we save a password hash ?'''
        return app_settings.KEEP_PASSWORD

    def service_principal(self):
        '''Service principal for checking password'''
        if not app_settings.SERVICE_PRINCIPAL:
            raise ImproperlyConfigured('Kerberos password backend needs '
                    'the setting KERBEROS_SERVICE_PRINCIPAL to be '
                    'set')
        return app_settings.SERVICE_PRINCIPAL

    def authenticate(self, username=None, password=None, **kwargs):
        '''Verify username and password using Kerberos'''
        if not username:
            return

        principal = force_bytes(self.principal_from_username(username))
        password = force_bytes(password)

        try:
            if not kerberos.checkPassword(principal, password,
                    self.service_principal(), self.default_realm()):
                return
        except kerberos.KrbError, e:
            logging.getLogger(__name__).error('password validation'
                    'for principal %r failed %s', principal, e)
            return
        else:
            if principal and self.authorize_principal(principal):
                user = self.lookup_user(principal)
                if self.keep_password():
                    user.set_password(password)
                return user



import os.path

INSTALLED_APPS = (
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django.contrib.sessions',
    'django_kerberos',
)
TEMPLATE_DIRS = (os.path.join(os.path.dirname(__file__), 'templates'),)
ROOT_URLCONF = 'django_kerberos.urls'
SECRET_KEY = 'xxx'



from django.conf.urls import patterns, include, url

from django.contrib import admin
admin.autodiscover()

from . import views

urlpatterns = patterns('',
    # Examples:
    url(r'^$', views.home, name='home'),
    # url(r'^blog/', include('blog.urls')),

    url('^accounts/kerberos/', include('django_kerberos.urls')),
    url('^accounts/', include('django.contrib.auth.urls')),
    url(r'^admin/', include(admin.site.urls)),
)



#! /usr/bin/env python

import subprocess
import os

from setuptools import setup, find_packages
from setuptools.command.sdist import sdist


class eo_sdist(sdist):
    def run(self):
        print "creating VERSION file"
        if os.path.exists('VERSION'):
            os.remove('VERSION')
        version = get_version()
        version_file = open('VERSION', 'w')
        version_file.write(version)
        version_file.close()
        sdist.run(self)
        print "removing VERSION file"
        if os.path.exists('VERSION'):
            os.remove('VERSION')


def get_version():
    '''Use the VERSION, if absent generates a version with git describe, if not
       tag exists, take 0.0.0- and add the length of the commit log.
    '''
    if os.path.exists('VERSION'):
        with open('VERSION', 'r') as v:
            return v.read()
    if os.path.exists('.git'):
        p = subprocess.Popen(['git', 'describe', '--dirty', '--match=v*'], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        result = p.communicate()[0]
        if p.returncode == 0:
            result = result.split()[0][1:]
        else:
            result = '0.0.0-%s' % len(subprocess.check_output(
                ['git', 'rev-list', 'HEAD']).splitlines())
        return result.replace('-', '.').replace('.g', '+g')
    return '0.0.0'

setup(name="django-kerberos",
      version=get_version(),
      license="AGPLv3 or later",
      description="Kerberos authentication for Django",
      long_description=file('README').read(),
      url="http://dev.entrouvert.org/projects/authentic/",
      author="Entr'ouvert",
      author_email="info@entrouvert.org",
      maintainer="Benjamin Dauvergne",
      maintainer_email="bdauvergne@entrouvert.com",
      packages=find_packages('src'),
      install_requires=[
          'django>1.5',
          'pykerberos',
      ],
      package_dir={
          '': 'src',
      },
      package_data={
          'django_kerberos': [
              'templates/django_kerberos/*.html',
              'static/js/*.js',
          ],
      },
      cmdclass={'sdist': eo_sdist})



import sys

class AppSettings(object):
    __PREFIX = 'KERBEROS_'
    __DEFAULTS = {
            'BACKEND_CREATE': False,
            'BACKEND_ADMIN_REGEXP': None,
            'DEFAULT_REALM': None,
            'SERVICE_PRINCIPAL': '',
            'HOSTNAME': None,
            'KEEP_PASSWORD': False,
    }

    def __getattr__(self, name):
        from django.conf import settings
        if name not in self.__DEFAULTS:
            raise AttributeError
        return getattr(settings, self.__PREFIX + name, self.__DEFAULTS[name])

app_settings = AppSettings()
app_settings.__name__ = __name__
sys.modules[__name__] = app_settings
